/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2012 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */

package org.glassfish.osgicdi.impl;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.logging.Logger;

import javax.enterprise.context.spi.CreationalContext;
import javax.enterprise.inject.Any;
import javax.enterprise.inject.Default;
import javax.enterprise.inject.spi.Bean;
import javax.enterprise.inject.spi.BeanManager;
import javax.enterprise.util.Nonbinding;
import javax.inject.Qualifier;

import org.glassfish.osgicdi.Publish;
import org.glassfish.osgicdi.ServiceProperty;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleReference;

/**
 * This class obtains @Publish class instance of the current bean bundle and
 * publishes it as OSGi services all that are {@link Publish} annotated.
 * <p/>
 * 
 * @author Tang Yong (tangyong@cn.fujitsu.com)
 */
public class OSGiServicePublisher {

	private static Logger logger = Logger.getLogger(OSGiServicePublisher.class
			.getPackage().getName());

	private List<Class<?>> classes;

	private List<BeanManager> bms;

	public OSGiServicePublisher(List<Class<?>> classes, List<BeanManager> bms) {
		this.classes = classes;
		this.bms = bms;
	}

	public void registerOSGiService() {
		for (Class<?> clazz : classes) {
			List<Annotation> qualifiers = getQualifiers(clazz);
			Object service = null;
			for (BeanManager bm : bms) {
				try {
					service = getContextualInstance(
							bm,
							clazz,
							qualifiers.toArray(new Annotation[qualifiers.size()]));
					if (service != null) {
						break;
					}
				} catch (Throwable t) {
					t.printStackTrace();
				}
			}

			publish(clazz, service, qualifiers);
		}
	}

	/**
	 * Get a single CDI managed instance of a specific class. Return only the
	 * first result if multiple beans are available.
	 * <p/>
	 * <b>NOTE:</b> Using this method should be avoided if container provided
	 * injection is available.
	 * 
	 * @param manager
	 *            The bean manager with which to perform the lookup.
	 * @param type
	 *            The class for which to return an instance.
	 * @param qualifiers
	 *            qualifiers for current bean to perform the lookup.
	 * @return The managed instance, or null if none could be provided.
	 */
	@SuppressWarnings("unchecked")
	private <T> T getContextualInstance(BeanManager manager, Class<T> type,
			Annotation... qualifiers) {
		T result = null;
		Bean<T> bean = null;
		if (qualifiers.length == 0) {
			bean = (Bean<T>) manager.resolve(manager.getBeans(type));
		} else {
			bean = (Bean<T>) manager
					.resolve(manager.getBeans(type, qualifiers));
		}

		if (bean != null) {
			CreationalContext<T> context = manager
					.createCreationalContext(bean);
			if (context != null) {
				result = (T) manager.getReference(bean, type, context);
			}
		}
		return result;
	}

	private List<Annotation> getQualifiers(Class<?> clazz) {
		List<Annotation> qualifiers = new ArrayList<Annotation>();
		for (Annotation annotation : clazz.getAnnotations()) {
			if (annotation.annotationType()
					.isAnnotationPresent(Qualifier.class)) {
				qualifiers.add(annotation);
			}
		}
		return qualifiers;
	}

	/**
	 * This method publish a bean class as an OSGi service.
	 * <p/>
	 * The class is published under (in order):
	 * <ul>
	 * <li>The contract list of its @Publish annotation
	 * <li>Its interfaces
	 * <li>Its super type
	 * <li>Itself implementation class
	 * </ul>
	 * with the right properties extracted from its annotations.
	 * 
	 * @param clazz
	 * @param service
	 * @param qualifiers
	 */
	private void publish(Class<?> clazz, Object service,
			List<Annotation> qualifiers) {
		Publish publish = clazz.getAnnotation(Publish.class);
		Class<?>[] contracts = publish.contracts();
		ServiceProperty[] serverProps = publish.serviceProperties();

		Properties properties = getServiceProperties(qualifiers, serverProps);

		BundleContext ctx = BundleReference.class.cast(clazz.getClassLoader())
				.getBundle().getBundleContext();

		if (contracts.length > 0) {
			// if there are contracts and these contracts are valid
			String[] names = new String[contracts.length];
			for (int i = 0; i < contracts.length; i++) {
				if (contracts[i].isAssignableFrom(clazz)) {
					names[i] = contracts[i].getName();
				} else {
					RuntimeException e = new RuntimeException("Contract "
							+ contracts[i] + " is not assignable from " + clazz
							+ ", or is not an interface."
							+ " Unable to publish the service " + clazz);

					throw e;
				}
			}

			ctx.registerService(names, service, properties);
		} else {
			if (clazz.getInterfaces().length > 0) {
				// if there are multiple interfaces
				List<Class<?>> interfaces = new ArrayList<Class<?>>();
				for (Class<?> itf : clazz.getInterfaces()) {
					interfaces.add(itf);
				}

				String[] names = new String[interfaces.size()];
				for (int i = 0; i < interfaces.size(); i++) {
					names[i] = interfaces.get(i).getName();
				}

				ctx.registerService(names, service, properties);

			} else {
				Class<?> superClass = clazz.getClass().getSuperclass();
				if (superClass != null && superClass != Object.class) {
					// if there is a superclass and the superclass is not equal
					// Object
					ctx.registerService(superClass.getName(), service,
							properties);
				} else {
					// register the implementation type itself directly
					ctx.registerService(clazz.getName(), service, properties);
				}
			}
		}
	}

	/**
	 * This method obtains OSGi service properties for @Publish classes.
	 * <p/>
	 * obtaining rules are as following:
	 * <ul>
	 * <li>getting @Publish's ServiceProperty[]
	 * <li>if qualifiers exist on @Publish classes, scanning these qualifiers,
	 * <li>if current scanned qualifier is not @Default or @Any, we do the
	 * following,
	 * <li>
	 * <p>
	 * 1) if the qualifier does not contain any member(or method), using the
	 * qualifier's simple name as service property key, '*' as service property
	 * value. For example, a user defines @Blue qualifier, and the qualifier
	 * does not contain any method, then, service property will be defined as
	 * (blue=*) .
	 * </p>
	 * <li>
	 * <p>
	 * 2) if the qualifier contains some members(or methods) , scanning these
	 * methods, once finding current method is not a Nonbinding, using the
	 * qualifier's simple name + '.' + the method name as service property key,
	 * the method's return value as service property value. for example, a user
	 * defines @Blue(name="BlueColor"), service property will be defined as
	 * (blue.name=BlueColor)
	 * </ul>
	 * 
	 * @param qualifiers
	 *            qualifiers on @Publish classes
	 * @param serverProps
	 *            @Publish's ServiceProperty[]
	 * @return Properties
	 */
	private Properties getServiceProperties(List<Annotation> qualifiers,
			ServiceProperty[] serverProps) {
		Properties properties = null;

		if (serverProps.length > 0) {
			properties = new Properties();

			for (ServiceProperty serviceProperty : serverProps) {
				properties.setProperty(serviceProperty.name(),
						serviceProperty.value());
			}
		} else {
			if (!qualifiers.isEmpty()) {
				properties = new Properties();

				for (Annotation qualifier : qualifiers) {
					if (!qualifier.annotationType().equals(Default.class)
							&& !qualifier.annotationType().equals(Any.class)) {
						if (qualifier.annotationType().getDeclaredMethods().length > 0) {
							for (Method m : qualifier.annotationType()
									.getDeclaredMethods()) {
								if (!m.isAnnotationPresent(Nonbinding.class)) {
									try {
										String key = qualifier.annotationType()
												.getSimpleName().toLowerCase()
												+ "."
												+ m.getName().toLowerCase();
										Object value = m.invoke(qualifier);
										if (value == null) {
											value = m.getDefaultValue();
											if (value == null) {
												value = "*";
											}
										}
										properties.setProperty(key,
												value.toString());
									} catch (Throwable t) {// ignore
									}
								}
							}
						} else {
							properties.setProperty(qualifier.annotationType()
									.getSimpleName().toLowerCase(), "*");
						}
					}
				}
			}
		}

		return properties;
	}
}
